%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% cost sensitive support vector machine solution path along cost asymmetry
%
% gyemin lee    2007/08/13
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [gamma alpha alpha0] = ...
	cssvm_path_gamma(x, y, hkernel, kernelparam, lambda, ...
					 init_alpha, init_alpha0, init_elbow, btrace)
% function [gamma alpha alpha0] = ...
% 	cssvm_path_gamma(x, y, hkernel, kernelparam, lambda, ...
% 					 init_alpha, init_alpha0, init_elbow, btrace)
%	x				N x p feature vector
%	y				N x 1 label
%	hkernel			kernel function handle
%	kernelparam		kernel parameter
%	lambda			regularization parameter
%	init_alpha		alpha for the balanced svm (gamma=n-/n)
%	init_alpha0		alpha0
%	init_elbow		elbow
%	btrace			boolean trace
%
if nargin < 9, btrace = true; end

eps = 1e-10;
gammamin = eps;
gammamax = 1;

N = length(y);
Np = sum(y== 1);
Nm = sum(y==-1);
init_gamma = Nm/N;	% balanced svm

% output variables
nmoves = 5*N;				% maximum number of steps
gamma = zeros(nmoves,1);
alpha = zeros(N,nmoves);
alpha0 = zeros(nmoves, 1);
stats = zeros(nmoves,3);

K = hkernel(x, x, kernelparam);		% kernel matrix
Kt = diag(y)*K*diag(y);

% initialize
k = 1;		% step
gamma(k) = init_gamma;
alpha(:,k) = init_alpha;
alpha0(k) = init_alpha0;

fl = (K*(alpha(:,k).*y) + alpha0(k)) / lambda;

% initialize active sets
[Right Elbow Left] = svminit(init_alpha, init_elbow);

obs = Elbow;
movefrom = repmat(' ', 1, length(Elbow));
moveto   = repmat('E', 1, length(Elbow));

stats(k,:) = svmstat(y, fl, Elbow);

if (btrace)
	strtrace = svmprint(k, obs, movefrom, moveto, gamma(k), stats(k,:));
end

% decrease gamma to 0
gammaout = -1;
while (k<nmoves && gamma(k)>gammamin)
	if (isempty(Elbow))
		% the elbow is empty. need to reinitialize.
		[flmax idx] = max(fl(Left).*y(Left));
		alpha0t = y(Left(idx))*lambda - K(Left(idx),:)*(alpha(:,k).*y);
		Elbowt = Left(idx);
		
		gamma(k+1) = gamma(k);
		alpha(:,k+1) = alpha(:,k);
		alpha0(k+1) = alpha0t;
		Elbow = Elbowt;
		Left = setdiff(Left, Elbow);
		
% 		fl = (K*(alpha(:,k+1).*y) + alpha0(k+1)) / lambda;
		fl = fl + (alpha0(k+1)-alpha0(k))/lambda;
		
		obs = Elbow;
		movefrom=' ';	moveto='E';
	else
		Kstar = [0 y(Elbow)'; y(Elbow) Kt(Elbow, Elbow)];
		bstar = Kstar\([length(Left); y(Elbow).*sum(K(Elbow, Left),2)]);
		b0 = bstar(1);
		b = bstar(2:end);
		
		hl = K(:,Elbow)*(b.*y(Elbow)) + b0 - sum(K(:,Left),2);
		
		% check for exits from elbow
		temp = ~(abs(1+y(Elbow).*b)<eps);
		if (length(temp)==1 && temp == 0), gammaleft = gammaout; 
		else
		gammaleft = zeros(size(temp));
		gammaleft(temp) = ((y(Elbow(temp))==-1)+y(Elbow(temp)).*(alpha(Elbow(temp),k)+gamma(k)*b(temp)))./(1+y(Elbow(temp)).*b(temp));
		gammaleft(~temp) = gammaout;
		end
		
		temp = ~(abs(b)<eps);
		if (length(temp)==1 && temp == 0), gammaright = gammaout; 
		else
		gammaright = zeros(size(temp));
		gammaright(temp) = gamma(k)+alpha(Elbow(temp),k)./b(temp);
		gammaright(~temp) = gammaout;
		end
		
		gammarl = [gammaright; gammaleft];
		gammaexit = max(gammarl(gammarl < gamma(k) - eps));
		if isempty(gammaexit), gammaexit = gammaout; end
		
		% check for entry into elbow
		temp = ~(abs(hl)<eps);
		temp(Elbow) = false;
		gammae = zeros(size(temp));
		gammae(temp) = gamma(k)+lambda*(fl(temp)-y(temp))./hl(temp);
		gammae(~temp) = gammaout;
		gammaentry = max(gammae(gammae < gamma(k) - eps));
		if isempty(gammaentry), gammaentry = gammaout; end
		
		gammanext = max(gammaexit, gammaentry);
		if (gammanext<gammamin)
			gamma(k+1) = 0;
			alpha(:,k+1) = 0;
			alpha0(k+1) = -lambda;
			fl(:) = -1;
			
			k = k + 1;

			stats(k,:) = svmstat(y, fl, Elbow);

			if (btrace)
				strtrace = [strtrace, ...
					sprintf('%3d:              gamma=%6.6f             FP=%3d  FN=%3d\n', ...
						k, gamma(k), stats(k,2), stats(k,3))];
			end
			break
		end
		
		% update gamma, alpha
		gamma(k+1) = gammanext;
% 		alpha(:,k+1) = alpha(:,k);
		alpha(Right,k+1) = 0;
		alpha(Left,k+1) = (y(Left)==-1)+y(Left)*gammanext;
		alpha(Elbow,k+1) = alpha(Elbow,k) - (gammanext-gamma(k))*b;
		alpha0(k+1) = alpha0(k) - (gammanext-gamma(k))*b0;
		fl = (gamma(k)-gammanext)/lambda*hl+fl;
		
		% update sets
		if (gammaentry > gammaexit)
			iin = find(gammae==gammaentry);
			obs = iin;
			
			tempe = ismember(iin, Left);
			movefrom = repmat('R', 1, length(iin));
			movefrom(tempe) = 'L';
			
			Left = setdiff(Left, iin);
			Right = setdiff(Right, iin);
			
			Elbow = [Elbow; iin];
			moveto = repmat('E', 1, length(iin));
		else
		    movefrom = 'E';
		    
		    tempr = abs(gammaright - gammaexit)<eps;
		    toright = Elbow(tempr);
		    templ = abs(gammaleft - gammaexit)<eps;
		    toleft = Elbow(templ);
		    temp = tempr | templ;
		    
		    Elbow = Elbow(~temp);
		    Right = [Right; toright];
		    moveto = repmat('R', 1, length(toright));
		    Left = [Left; toleft];
		    moveto = [moveto, repmat('L', 1, length(toleft))];
		
		    obs = [toright; toleft];
			movefrom = repmat('E', 1, length(moveto));
		end
	end
	
	k = k + 1;
	
	stats(k,:) = svmstat(y, fl, Elbow);
	
	if (btrace)
		strtrace = [strtrace, ...
			svmprint(k, obs, movefrom, moveto, gamma(k), stats(k,:))];
	end
end

% initialize
[gamma(1:k) ix] = sort(gamma(1:k),'ascend');
alpha(:,1:k) = alpha(:,ix);
alpha0(1:k) = alpha0(ix);

fl = (K*(alpha(:,k).*y) + alpha0(k)) / lambda;

% initialize active sets
[Right Elbow Left] = svminit(init_alpha, init_elbow);

% increase gamma to 1
gammaout = 2;
while (k<nmoves && gamma(k)<gammamax)
	if (isempty(Elbow))
		% the elbow is empty. need to reinitialize.
		[flmax idx] = max(fl(Left).*y(Left));
		alpha0t = y(Left(idx))*lambda - K(Left(idx),:)*(alpha(:,k).*y);
		Elbowt = Left(idx);
		
		gamma(k+1) = gamma(k);
		alpha(:,k+1) = alpha(:,k);
		alpha0(k+1) = alpha0t;
		Elbow = Elbowt;
		Left = setdiff(Left, Elbow);
		
% 		fl = (K*(alpha(:,k+1).*y) + alpha0(k+1)) / lambda;
		fl = fl + (alpha0(k+1)-alpha0(k))/lambda;
		
		obs = Elbow;
		movefrom=' ';	moveto='E';
	else
		Kstar = [0 y(Elbow)'; y(Elbow) Kt(Elbow, Elbow)];
		bstar = Kstar\([length(Left); y(Elbow).*sum(K(Elbow, Left),2)]);
		b0 = bstar(1);
		b = bstar(2:end);
		
		hl = K(:,Elbow)*(b.*y(Elbow)) + b0 - sum(K(:,Left),2);
		
		% check for exits from elbow
		temp = ~(abs(1+y(Elbow).*b)<eps);
		if (length(temp)==1 && temp == 0), gammaleft = gammaout; 
		else
		gammaleft = zeros(size(temp));
		gammaleft(temp) = ((y(Elbow(temp))==-1)+y(Elbow(temp)).*(alpha(Elbow(temp),k)+gamma(k)*b(temp)))./(1+y(Elbow(temp)).*b(temp));
		gammaleft(~temp) = gammaout;
		end
		
		temp = ~(abs(b)<eps);
		if (length(temp)==1 && temp == 0), gammaright = gammaout; 
		else
		gammaright = zeros(size(temp));
		gammaright(temp) = gamma(k)+alpha(Elbow(temp),k)./b(temp);
		gammaright(~temp) = gammaout;
		end
		
		gammarl = [gammaright; gammaleft];
		gammaexit = min(gammarl(gammarl > gamma(k) + eps));
		if isempty(gammaexit), gammaexit = gammaout; end
		
		% check for entry into elbow
		temp = ~(abs(hl)<eps);
		temp(Elbow) = false;
		gammae = zeros(size(temp));
		gammae(temp) = gamma(k)+lambda*(fl(temp)-y(temp))./hl(temp);
		gammae(~temp) = gammaout;
		gammaentry = min(gammae(gammae > gamma(k) + eps));
		if isempty(gammaentry), gammaentry = gammaout; end
		
		gammanext = min(gammaexit, gammaentry);
		if (gammanext>gammamax)
			gamma(k+1) = 1;
			alpha(:,k+1) = 0;
			alpha0(k+1) = lambda;
			fl(:) = 1;
			
			k = k + 1;

			stats(k,:) = svmstat(y, fl, Elbow);

			if (btrace)
				strtrace = [strtrace, ...
					sprintf('%3d:              gamma=%6.6f             FP=%3d  FN=%3d\n', ...
						k, gamma(k), stats(k,2), stats(k,3))];
			end
			break
		end
		
		% update gamma, alpha
		gamma(k+1) = gammanext;
% 		alpha(:,k+1) = alpha(:,k);
		alpha(Right,k+1) = 0;
		alpha(Left,k+1) = (y(Left)==-1)+y(Left)*gammanext;
		alpha(Elbow,k+1) = alpha(Elbow,k) - (gammanext-gamma(k))*b;
		alpha0(k+1) = alpha0(k) - (gammanext-gamma(k))*b0;
		fl = (gamma(k)-gammanext)/lambda*hl+fl;
		
		% update sets
		if (gammaentry < gammaexit)
			iin = find(gammae==gammaentry);
			obs = iin;
			
			tempe = ismember(iin, Left);
			movefrom = repmat('R', 1, length(iin));
			movefrom(tempe) = 'L';
			
			Left = setdiff(Left, iin);
			Right = setdiff(Right, iin);
			
			Elbow = [Elbow; iin];
			moveto = repmat('E', 1, length(iin));
		else
		    movefrom = 'E';
		    
		    tempr = abs(gammaright - gammaexit)<eps;
		    toright = Elbow(tempr);
		    templ = abs(gammaleft - gammaexit)<eps;
		    toleft = Elbow(templ);
		    temp = tempr | templ;
		    
		    Elbow = Elbow(~temp);
		    Right = [Right; toright];
		    moveto = repmat('R', 1, length(toright));
		    Left = [Left; toleft];
		    moveto = [moveto, repmat('L', 1, length(toleft))];
		
		    obs = [toright; toleft];
			movefrom = repmat('E', 1, length(moveto));
		end
	end
	
	k = k + 1;
	
	stats(k,:) = svmstat(y, fl, Elbow);
	
	if (btrace)
		strtrace = [strtrace, ...
			svmprint(k, obs, movefrom, moveto, gamma(k), stats(k,:))];
	end
end

% outputs
gamma = gamma(1:k);
alpha = alpha(:,1:k);
alpha0 = alpha0(1:k);
stats = stats(1:k,:);

% print the result
if (btrace)
	fresult = fopen('gamma.txt', 'w');
	fwrite(fresult, strtrace);
	fclose(fresult);
%	strtrace
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [Right Elbow Left] = svminit(init_alpha, init_elbow)
% function [Right Elbow Left] = svminit(init_alpha, init_elbow)
%	initialize three active sets: Elbow, Right, and Left
% 
N = length(init_alpha);

Elbow = init_elbow;
temp = init_alpha<eps;
temp(Elbow) = false;
Right = find(temp);
Left  = setdiff((1:N), union(Right, Elbow))';


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [alpha0 Elbow] = svmreinit(y, K, lambda, alpha, fl)
% function [gamma alpha alpha0 Elbow] = svmreinit(y, K, lambda, alpha, fl)
%	reinitialize when Elbow is empty
%
[flmax idx] = max(fl.*y);
alpha0 = y(idx)*lambda - K(idx,:)*(alpha.*y);
Elbow = idx;

%{
% function [gamma alpha alpha0 Elbow] = svminit(y, K, lambda)
% method1 -> failed!!
np = sum(K(y== 1,:),2);
nm = sum(K(y==-1,:),2);
bp = sum(K(y== 1,y==-1),2);
bm = sum(K(y==-1,y==-1),2);

gamma = sum(y==-1)/length(y);
rmin = gamma;
alpha = gamma*(y==1) + (1-gamma)*(y==-1);

gp = K(y==1,:)*(alpha.*y);
gm = K(y==-1,:)*(alpha.*y);

ip = find(gp==max(gp));
im = find(gm==min(gm));

while(1)
	rp = (bp(ip)-bp)./(np(ip)-np);
	rm = (bm(im)-bm)./(nm(im)-nm);

	rmp = min(rp(rp>rmin));
	rmm = min(rm(rm>rmin));
	rmax = min(rmp, rmm);
	nip = find(rp==rmp);
	nim = find(rm==rmm);

	rnext = (2*lambda+bp(ip)-bm(im))./(np(ip)-nm(im));
	if (rnext>rmin && rnext<rmax)
		gamma = rnext;
		alpha = gamma*(y==1) + (1-gamma)*(y==-1);
		alpha0 = lambda - gamma*np(ip)+bp(ip);
		Elbow = [ip; im];
		break
	end

	rmin = rmax;
	if (rmp<rmm)
		ip = find(rp==rmp);
	else
		im = find(rm==rmm);
	end
end
%}


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function stats = svmstat(y, f, Elbow)
% function stats = svmstat(y, f, Elbow)
%	statistics
%	size of Elbow, # of false positive, # of false negative
%
stats = zeros(1,3);
stats(1) = length(Elbow);		% size of Elbow
stats(2) = sum(f(y==-1)>0);		% false positive
stats(3) = sum(f(y== 1)<0);		% false negative


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function str = svmprint(step, obs, movefrom, moveto, gamma, stats)
% function str = svmprint(step, obs, movefrom, moveto, gamma, stats)
%	trace the solution path
%
str = '';
for ii=1:length(obs)
	strtemp = sprintf('%3d: Obs %3d %c->%c gamma=%6.6f  Elbow=%3d  FP=%3d  FN=%3d\n', ...
		step, obs(ii), movefrom(ii), moveto(ii), gamma, stats(1), stats(2), stats(3));
	str = [str, strtemp];
end
