%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% cost sensitive support vector machine solution path along cost asymmetry
%
% gyemin lee    2007/08/13
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [gamma alpha] = ...
	cssvm_path_gamma_wb(x, y, hkernel, kernelparam, lambda, ...
					 init_alpha, init_elbow, btrace)
% function [gamma alpha] = ...
% 	cssvm_path_gamma_wb(x, y, hkernel, kernelparam, lambda, ...
% 					 init_alpha, init_elbow, btrace)
%	x				N x p feature vector
%	y				N x 1 label
%	hkernel			kernel function handle
%	kernelparam		kernel parameter
%	lambda			regularization parameter
%	init_alpha		alpha for the balanced svm (gamma=n-/n)
%	init_elbow		elbow
%	btrace			boolean trace
%
if nargin < 8, btrace = false; end

EPS = 1e-10;
EPS2 = 1e-14;
gammamin = EPS;
gammamax = 1;

N = length(y);
Np = sum(y== 1);
Nm = sum(y==-1);
init_gamma = Nm/N;	% balanced svm

% output variables
nmoves = 5*N;				% maximum number of steps
gamma = zeros(nmoves,1);
alpha = zeros(N,nmoves);
stats = zeros(nmoves,3);

K = hkernel(x, x, kernelparam);		% kernel matrix
Kt = diag(y)*K*diag(y);

% initialize
k = 1;		% step
gamma(k) = init_gamma;
alpha(:,k) = init_alpha;

fl = (K*(alpha(:,k).*y)) / lambda;

% initialize active sets
% [Right Elbow Left] = svminit(y, init_gamma, init_alpha, init_elbow, y.*fl, false);
[Right Elbow Left] = svminit(y, Kt, lambda, init_gamma, init_alpha, init_elbow, false);

iin = [];
obs = Elbow;
movefrom = repmat(' ', 1, length(Elbow));
moveto   = repmat('E', 1, length(Elbow));

stats(k,:) = svmstat(y, fl, Elbow);

if (btrace)
	strtrace = svmprint(k, obs, movefrom, moveto, gamma(k), stats(k,:));
end

% decrease gamma to 0
gammaout = -1;
while (gamma(k)>gammamin)
	if (isempty(Elbow))
% 		% the elbow is empty. need to reinitialize.
% 		[gammat alphat Elbowt] = svmreinit(y, K, lambda, gamma(k), Left, Right, false);
		[gammat alphat Elbowt] = svmreinit(y, K, lambda, Left, Right, false);
        if (isempty(gammat) || gammat<0)
            break
        end
        if (gamma(k)<gammat)
            fprintf('WARNING: CS-SVM : empty Elbow dec %d\n', k);
            break
        end
		gamma(k+1) = gammat;
		alpha(:,k+1) = alphat;
		Elbow = Elbowt;
		Left = setdiff(Left, Elbow);
		Right = setdiff(Right, Elbow);
		
		fl = (K*(alpha(:,k+1).*y)) / lambda;
		
		obs = Elbow;
		movefrom = ' ';		moveto = 'E';
    else
		Kstar = Kt(Elbow, Elbow);
		b = Kstar\(y(Elbow).*sum(K(Elbow, Left),2));
		
		hl = K(:,Elbow)*(b.*y(Elbow)) - sum(K(:,Left),2);
		
		% check for exits from elbow
		temp = ~(abs(1+y(Elbow).*b)<EPS);
		if (length(temp)==1 && temp == 0), gammaleft = gammaout; 
		else
		gammaleft = zeros(size(temp));
		gammaleft(temp) = ((y(Elbow(temp))==-1)+y(Elbow(temp)).*(alpha(Elbow(temp),k)+gamma(k)*b(temp)))./(1+y(Elbow(temp)).*b(temp));
		gammaleft(~temp) = gammaout;
		end
		
		temp = ~(abs(b)<EPS);
		if (length(temp)==1 && temp == 0), gammaright = gammaout; 
		else
		gammaright = zeros(size(temp));
		gammaright(temp) = gamma(k)+alpha(Elbow(temp),k)./b(temp);
		gammaright(~temp) = gammaout;
		end
		
		gammarl = [gammaright; gammaleft];
		gammaexit = max(gammarl(gammarl < gamma(k) - EPS2));
		if isempty(gammaexit), gammaexit = gammaout; end
		
		% check for entry into elbow
		temp = ~(abs(hl)<EPS);
		temp(Elbow) = false;
        temp(obs) = false;
		gammae = zeros(size(temp));
		gammae(temp) = gamma(k)+lambda*(fl(temp)-y(temp))./hl(temp);
		gammae(~temp) = gammaout;
		gammaentry = max(gammae(gammae < gamma(k) - EPS2));
		if isempty(gammaentry), gammaentry = gammaout; end
		
		gammanext = max(gammaexit, gammaentry);
		if (gammanext<gammamin)
            gammanext = 0;
            gamma(k+1) = gammanext;
            alpha(Right,k+1) = 0;
            alpha(Left,k+1) = (y(Left)==-1)+y(Left)*gammanext;
            alpha(Elbow,k+1) = alpha(Elbow,k) - (gammanext-gamma(k))*b;
            k=k+1;
            stats(k,:) = svmstat(y, fl, Elbow);
			break
		end
		
		% update gamma, alpha
		gamma(k+1) = gammanext;
		alpha(Right,k+1) = 0;
		alpha(Left,k+1) = (y(Left)==-1)+y(Left)*gammanext;
		alpha(Elbow,k+1) = alpha(Elbow,k) - (gammanext-gamma(k))*b;
		fl = (gamma(k)-gammanext)/lambda*hl+fl;
		
		% update sets
		if (gammaentry > gammaexit)
% 			iin = find(gammae==gammaentry);
            iin = find(abs(gammae-gammaentry)<EPS);
			obs = iin;
			
			tempe = ismember(iin, Left);
			movefrom = repmat('R', 1, length(iin));
			movefrom(tempe) = 'L';
			
			Left = setdiff(Left, iin);
			Right = setdiff(Right, iin);
			
			Elbow = [Elbow; iin];
			moveto = repmat('E', 1, length(iin));
		else
		    movefrom = 'E';
		    
		    tempr = abs(gammaright - gammaexit)<EPS;
		    toright = Elbow(tempr);
		    templ = abs(gammaleft - gammaexit)<EPS;
		    toleft = Elbow(templ);
		    temp = tempr | templ;
		    
		    Elbow = Elbow(~temp);
		    Right = [Right; toright];
		    moveto = repmat('R', 1, length(toright));
		    Left = [Left; toleft];
		    moveto = [moveto, repmat('L', 1, length(toleft))];
		
		    obs = [toright; toleft];
			movefrom = repmat('E', 1, length(moveto));
		end
	end
	
	k = k + 1;
	
	stats(k,:) = svmstat(y, fl, Elbow);
	
	if (btrace)
		strtrace = [strtrace, ...
			svmprint(k, obs, movefrom, moveto, gamma(k), stats(k,:))];
	end
end

if (btrace)
    strtrace = [strtrace, sprintf('----------\n')];
end

% initialize
[gamma(1:k) ix] = sort(gamma(1:k),'ascend');
alpha(:,1:k) = alpha(:,ix);

fl = (K*(alpha(:,k).*y)) / lambda;

% initialize active sets
% [Right Elbow Left] = svminit(y, init_gamma, init_alpha, init_elbow, y.*fl, true);
[Right Elbow Left] = svminit(y, Kt, lambda, init_gamma, init_alpha, init_elbow, true);

obs = [];

% increase gamma to 1
gammaout = 2;
while (gamma(k)<gammamax)
	if (isempty(Elbow))
% 		% the elbow is empty. need to reinitialize.
% 		[gammat alphat Elbowt] = svmreinit(y, K, lambda, gamma(k), Left, Right, true);
		[gammat alphat Elbowt] = svmreinit(y, K, lambda, Left, Right, true);
        if (isempty(gammat) || gammat>1)
            break
        end
        if (gamma(k)>gammat)
            fprintf('WARNING: CS-SVM : empty Elbow inc %d\n', k);
            break
        end
		gamma(k+1) = gammat;
		alpha(:,k+1) = alphat;
		Elbow = Elbowt;
		Left = setdiff(Left, Elbow);
		Right = setdiff(Right, Elbow);
		
		fl = (K*(alpha(:,k+1).*y)) / lambda;
		
		obs = Elbow;
		movefrom = ' ';		moveto = 'E';
	else
		Kstar = Kt(Elbow, Elbow);
		b = Kstar\(y(Elbow).*sum(K(Elbow, Left),2));
		
		hl = K(:,Elbow)*(b.*y(Elbow)) - sum(K(:,Left),2);
		
		% check for exits from elbow
		temp = ~(abs(1+y(Elbow).*b)<EPS);
		if (length(temp)==1 && temp == 0), gammaleft = gammaout; 
		else
		gammaleft = zeros(size(temp));
		gammaleft(temp) = ((y(Elbow(temp))==-1)+y(Elbow(temp)).*(alpha(Elbow(temp),k)+gamma(k)*b(temp)))./(1+y(Elbow(temp)).*b(temp));
		gammaleft(~temp) = gammaout;
		end
		
		temp = ~(abs(b)<EPS);
		if (length(temp)==1 && temp == 0), gammaright = gammaout; 
		else
		gammaright = zeros(size(temp));
		gammaright(temp) = gamma(k)+alpha(Elbow(temp),k)./b(temp);
		gammaright(~temp) = gammaout;
		end
		
		gammarl = [gammaright; gammaleft];
		gammaexit = min(gammarl(gammarl > gamma(k) + EPS2));
		if isempty(gammaexit), gammaexit = gammaout; end
		
		% check for entry into elbow
		temp = ~(abs(hl)<EPS);
		temp(Elbow) = false;
        temp(obs) = false;
		gammae = zeros(size(temp));
		gammae(temp) = gamma(k)+lambda*(fl(temp)-y(temp))./hl(temp);
		gammae(~temp) = gammaout;
		gammaentry = min(gammae(gammae > gamma(k) + EPS2));
		if isempty(gammaentry), gammaentry = gammaout; end
		
		gammanext = min(gammaexit, gammaentry);
		if (gammanext>gammamax)
            gammanext = 1;
            gamma(k+1) = gammanext;
            alpha(Right,k+1) = 0;
            alpha(Left,k+1) = (y(Left)==-1)+y(Left)*gammanext;
            alpha(Elbow,k+1) = alpha(Elbow,k) - (gammanext-gamma(k))*b;
            k=k+1;
            stats(k,:) = svmstat(y, fl, Elbow);
			break
		end
		
		% update gamma, alpha
		gamma(k+1) = gammanext;
		alpha(Right,k+1) = 0;
		alpha(Left,k+1) = (y(Left)==-1)+y(Left)*gammanext;
		alpha(Elbow,k+1) = alpha(Elbow,k) - (gammanext-gamma(k))*b;
		fl = (gamma(k)-gammanext)/lambda*hl+fl;
		
		% update sets
		if (gammaentry < gammaexit)
% 			iin = find(gammae==gammaentry);
            iin = find(abs(gammae-gammaentry)<EPS);
			obs = iin;
			
			tempe = ismember(iin, Left);
			movefrom = repmat('R', 1, length(iin));
			movefrom(tempe) = 'L';
			
			Left = setdiff(Left, iin);
			Right = setdiff(Right, iin);
			
			Elbow = [Elbow; iin];
			moveto = repmat('E', 1, length(iin));
		else
		    movefrom = 'E';
		    
		    tempr = abs(gammaright - gammaexit)<EPS;
		    toright = Elbow(tempr);
		    templ = abs(gammaleft - gammaexit)<EPS;
		    toleft = Elbow(templ);
		    temp = tempr | templ;
		    
		    Elbow = Elbow(~temp);
		    Right = [Right; toright];
		    moveto = repmat('R', 1, length(toright));
		    Left = [Left; toleft];
		    moveto = [moveto, repmat('L', 1, length(toleft))];
		
		    obs = [toright; toleft];
			movefrom = repmat('E', 1, length(moveto));
		end
	end
	
	k = k + 1;
	
	stats(k,:) = svmstat(y, fl, Elbow);
	
	if (btrace)
		strtrace = [strtrace, ...
			svmprint(k, obs, movefrom, moveto, gamma(k), stats(k,:))];
	end
end

% outputs
gamma = gamma(1:k);
alpha = alpha(:,1:k);
stats = stats(1:k,:);

% feasiblity check
if ~isempty(find(alpha < -EPS*2 | alpha > repmat(y==-1,1,k)+y*gamma' + 2*EPS,1));
    fprintf('WARNING: CS-SVM : infeasible solution!!\n');
end
if (gamma(1)>EPS || gamma(end)<1-EPS)
    fprintf('WARNING: CS-SVM : gamma= %e to gamma=%e \n', gamma(1), gamma(end));
end

% print the result
if (btrace)
	fresult = fopen('cssvm_gam.txt', 'w');
	fwrite(fresult, strtrace);
	fclose(fresult);
%	strtrace
end


function [R E L] = svminit(y, Kt, lambda, gamma, alpha, initE, bInc)
% function [R E L] = svminit(y, Kt, lambda, gamma, alpha, initE, bInc)
%	initialize three active sets: Elbow, Right, and Left
% 
EPS = 5e-10;
N = length(alpha);
E = initE;

fk = Kt*alpha/lambda;
tempE = find(abs(fk-1)<EPS | (alpha>EPS & alpha<(y==-1)+y*gamma-EPS));
idx = alpha<EPS;
idx(tempE) = false;
tempR = find(idx);
tempL = setdiff((1:N), union(tempE,tempR))';

candR = find(alpha < EPS);
idxER = find(ismember(tempE, candR));
tempER = tempE(idxER);

candL = find(alpha > gamma*(y==1) + (1-gamma)*(y==-1) - EPS);
idxEL = find(ismember(tempE, candL));
tempEL = tempE(idxEL);

Kstar = Kt(tempE, tempE);
b = Kstar\(Kt(tempE,tempL)*y(tempL));

if (~isempty(idxER))
	gammaL = ((y(tempER)==-1)+y(tempER).*(alpha(tempER)+gamma*b(idxER)))./(1+y(tempER).*b(idxER));
	idx1 = (gammaL>gamma & ~(gammaL>1 & y(tempER)==-1)) | (gammaL < 0 & y(tempER)==1);
	if (~bInc)
		E = setdiff(tempE, tempE(idxER(idx1)));
	else
		E = setdiff(tempE, tempE(idxER(~idx1)));
	end
end

if (~isempty(idxEL))
	gammaR = gamma+alpha(tempEL)./b(idxEL);
	idx2 = (gammaR>gamma & ~(gammaR>1 & y(tempEL)==-1)) | (gammaR < 0 & y(tempEL)==1);
	if (~bInc)
		E = setdiff(tempE, tempE(idxEL(idx2)));
	else
		E = setdiff(tempE, tempE(idxEL(~idx2)));
	end
end

R = setdiff(candR, E);
L = setdiff((1:N), union(E,R))';

%{
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [gamma alpha Elbow] = svmreinit(y, K, lambda, gamma_old, L, R, bInc)
% function [gamma alpha Elbow] = svmreinit((y, K, lambda, gamma_old, L, R, bInc)
%	reinitialize when Elbow is empty
%
Lm = L(y(L)==-1);
gamma1 = (y*lambda + sum(K(:,Lm),2))./sum(K(:,L),2);

if (~bInc)
    gamma = max(gamma1(gamma1 < gamma_old));
else
    gamma = min(gamma1(gamma1 > gamma_old));
end

alpha = zeros(length(y),1);

if isempty(gamma)
    Elbow = [];
else
    Elbow = find(gamma == gamma1);
    alpha(L) = gamma*(y(L)==1) + (1-gamma)*(y(L)==-1);
end
%}

function [gamma alpha Elbow] = svmreinit(y, K, lambda, L, R, bInc)
% function [gamma alpha Elbow] = svmreinit((y, K, lambda, L, R, bInc)
%	reinitialize when Elbow is empty
%
EPS = 1e-10;

Lp = L(y(L)==1);
Lm = L(y(L)==-1);
Rp = R(y(R)==1);
Rm = R(y(R)==-1);

if (~bInc)
	temp1 = (-lambda+sum(K(Lm,Lm),2))./sum(K(Lm,L),2);
	temp2 = (lambda+sum(K(Rp,Lm),2))./sum(K(Rp,L),2);
	[gamma idx] = max([temp1; temp2]);
%     Elbow = [Lm(find(abs(temp1-gamma)<EPS)); Rp(find(abs(temp2-gamma)EPS))];
	Idx = [Lm; Rp];
	Elbow = Idx(idx);
    if (gamma<0)
        gamma=0;
        Elbow=[];
    end
else
	temp1 = (lambda+sum(K(Lp,Lm),2))./sum(K(Lp,L),2);
	temp2 = (-lambda+sum(K(Rm,Lm),2))./sum(K(Rm,L),2);
	[gamma idx] = min([temp1; temp2]);
%     Elbow = [Lm(find(abs(temp1-gamma)<EPS)); Rp(find(abs(temp2-gamma)EPS))];
	Idx = [Lp; Rm];
	Elbow = Idx(idx);
    if (gamma>1)
        gamma=1;
        Elbow=[];
    end
end
alpha = zeros(length(y),1);
alpha(L) = gamma*(y(L)==1) + (1-gamma)*(y(L)==-1);



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function stats = svmstat(y, f, Elbow)
% function stats = svmstat(y, f, Elbow)
%	statistics
%	size of Elbow, # of false positive, # of false negative
%
stats = zeros(1,3);
stats(1) = length(Elbow);		% size of Elbow
stats(2) = sum(f(y==-1)>=0);		% false positive
stats(3) = sum(f(y== 1)>=0);		% true positive


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function str = svmprint(step, obs, movefrom, moveto, gamma, stats)
% function str = svmprint(step, obs, movefrom, moveto, gamma, stats)
%	trace the solution path
%
str = '';
for ii=1:length(obs)
	strtemp = sprintf('%3d: Obs %3d %c->%c gamma=%6.6f  Elbow=%3d  FP=%3d  TP=%3d\n', ...
		step, obs(ii), movefrom(ii), moveto(ii), gamma, stats(1), stats(2), stats(3));
	str = [str, strtemp];
end
